(* Copyright (c) DeltaSoft - Croatia *)

unit SortList;

(*
This class (with it's source) is given as freeware,
with no guarantees. If it doesn't work, or puts your
computer to smoke, destroys your data or whatever,
DON'T BLAME ME !
Before using it in real life - TEST IT !
There's a test example "SortLst.dpr"
----------------------------------------------------

This class shows it's best performance with small lists (up to cca 5.000 records),
but it can be used for large lists too.
We have tested it on Cyrix P133+ (underclocked to 100 HMz) with 3 MB of free RAM,
with this structure:
	3 integers + 1 double field = 20 bytes, all key fields.

Results with 100.000 records:
*	1.) cca 160 sec to Create (and auto-sort) 100.000 records, (cca 2.5 MB of RAM allocated)
* 2.) cca 150 sec to Delete every second record (50.000 records deleted)
* 3.) cca 140 sec to Search for 100.000 records and create all that are missing (50.000 new records created, no new memory allocated)
* 4.) cca 1 sec (one second) to Search for 100.000 records (all 100.000 records found)

Results with 50.000 records:
*	1.) cca 36 sec to Create (and auto-sort) 50.000 records, (cca 1.3 MB of RAM allocated)
* 2.) cca 35 sec to Delete every second record (25.000 records deleted)
* 3.) cca 28 sec to Search for 50.000 records and create all that are missing (25.000 new records created, no new memory allocated)
* 4.) cca 1 sec (one second) to Search for 50.000 records (all 50.000 records found)

Results with 10.000 records:
*	1.) cca 1.4 sec to Create (and auto-sort) 10.000 records, (cca 250 KB of RAM allocated)
* 2.) cca 0.7 sec to Delete every second record (5.000 records deleted)
* 3.) cca 0.7 sec to Search for 10.000 records and create all that are missing (5.000 new records created, no new memory allocated)
* 4.) cca 0.2 sec to Search for 10.000 records (all 10.000 records found)

We used "SortLst.dpr" program for testing purposes,
compiled sa Console application.
----------------------------------------------------

TSortList can manage records of any size, but it has few limitations:
	1. Max index size (per record) is 255 bytes !
		- The sum of bytes occupied by key fields must not exceed 255 bytes
	2. Only this types of data can be used for key fields inside index:
		- Double	(can be any other float type of 8 bytes, like TDateTime)
		- LongInt	(can be any other integer type of 4 bytes)
		- Char 		(can be any other type of 1 byte, but it has to be converted to char for "FindKey")
		- Boolean	(not recommended - you never know which value will represent "True")
		- ShortString [ size ] (has to be the last key field, only one per index)
	3. Only one string field can be used as a key field for indexing,
		 and it has to be the last key field.
	4. For field that are not indexed, it doesn't matter that type you use.
		 For non-indexed fields all types are allowed.
	5. After posting a new record (Append - modify - Post), it is placed on his right place inside indexed list.
		 If you need to change the key fields afterwards (after the first posting), you will have to
		 Delete the record and create a new one, or the record will not be moved to new location inside a list.
	6. After a record is deleted, no memory is being freed. But, if you append new records to your list,
		 no new memory is needed until you exceed the amount of memory already allocated. This is because
		 new records are filled in on places of previously deleted records. This method speeds-up delete operations
		 and new insert operations (after delete), but if you use the list a long time,
		 it is usefull to "EmptyTable" if you don't need the data inside it.
		 After a call to "EmptyTable", all allocated memory is released.
	7. Sorting is performed on the block of fields (all kinds of fields), so
		 strings are not sorted alphabetically, and numbers are not sorted by value.
		 The purpose of sorting is to arange data for fast Find operation,
		 so you can check if a record with exact key fields already exists,
		 but it is not arranged (sorted) for viewing.
	8. If you move to another record before posting your modifications,
		 all modifications to the record will be canceled (discarded).
		 If you want to modify a record, you don't have to call "Edit" before.
		 You just modify it, and then "Post" it.
	9. When using FindKey, the values passed to it HAVE TO BE the same type
		 as the key values. That is, zero for double is 0.0, and for integer is 0.
		 Nil values are not allowed, because no type-conversation is made.
		 That is because the class doesn't see the record structure (definition),
		 so it can't make any type conversations.

----------------------------------------------------
For example, let's say we needed "MySortList" list,
structured like this (3 fields, first 2 indexed):
	Field1 of type integer * (this one is the key field)
	Field2 of type double *  (this is the second key field)
	Field3 of type double
-------------------------
* Here's how we would define it's structure (any names can be used) :)
	type
		TMySortList=class(TSortList)	// TMySortList is our List type
			Field1:integer;
			Field2,
			Field3:double;
			_:byte;	// The EndMark field (every structure has to have it)
			end;
	var
		MySortList:TMySortList;	// MySortList is our list variable
----------------
* Before you can use the list, you have to create (and define) it:
		MySortList:=TMySortList.Create;	// Created

* When created, you have to define it ("Field3" is the first non-indexed {non-key} field):
		with MySortList do Define(Field3,_);
* If you want it to be indexed by all fields, call "Define" like this:
		with MySortList do Define(_,_);
* And if you don't want it to be indexed ("Field1" is the first non-indexed field):
		with MySortList do Define(Field1,_);
-----------------

* Work with the list (look at the "SortLst.dpr" example)

------------------
* After you've done using it, you have to destroy it:
		MySortList.Destroy;

----------------------------------------------------
If you find this class usefull and you have some ideas on
increasing it's performance or adding additional feaatures that
wouldn't slower down the performance, please contact me.
My E-mail is <delta@deltasoft.hr>

:) Thanx

*)


interface

uses
	SysUtils;

type
	PSString=^ShortString;
	// Data structure
	TDataPtr=^TData;
	TData=array[0..MaxLongInt-1] of byte;
	// Index structure
	TListDataPtr=^TListData;
	TListData=array[1..MaxLongInt div 4] of integer;
	// This it the sorted list class
	TSortList=class
		private
		Temp:integer;
		FPtr:pointer;						// Points to the first variable
		MaxRecordCount,					// Last known maximal record count (allocated memory)
		FRecordCount:integer;		// Real table record count
		FRecNo:integer;					// Points to active record
		FIndex:TListDataPtr;		// Pointer to index table
		FData:TDataPtr;					// Pointer to data table
		FBoF,FEoF:boolean;			// at Begining Of File, at End Of File ?
		FISize,FISize2:byte;		// Index fields size (in bytes), deleted record mark
		FSize2,									// Size with additional index-length byte (for index comparisson)
		FSize:integer;					// Record fields size (in bytes -> includes index size)
		procedure ClearRecord;	// Clear all data - empty table
		procedure ReadRecord;		// Read data from memory
		procedure WriteRecord;	// Write data to memory
		procedure InsertRecord;	// Insert one empty record (before this one)
		function GoToNearest:boolean;	// Go to nearest record (matching index fields only), return TRUE if exact found
		function Find(const s:ShortString):boolean;	// Find best match

		public

		constructor Create; virtual;
		procedure Define(var NonIndex,EndMark);
		destructor Destroy; override;

		procedure EmptyTable;				// Make table Empty
		procedure Append;						// Append one record
		procedure Edit;							// Start editing data
		procedure Delete;						// Delete this record
		procedure First;						// To first record
		procedure Last;							// To last record
		procedure Next;							// Move one forward
		procedure Prior;						// Move one back
		procedure Move(a:integer);	// Move in any direction
		procedure Post;							// Save data to memory table
		procedure Cancel;						// Cancel editing or appending
		function FindKey(const Values: array of const):boolean;	// Search for a specific record (all key fields have to match)
		property BOF:boolean read FBof;	// Tried to pass the begining of "File" ?
		property EOF:boolean read FEof;	// Tried to pass the end of "File" ?
		property RecordCount:integer read FRecordCount;	// Number of records in a list
		property RecNo:integer read FRecNo;	// Active record number (starting from 1)
		end;

implementation

constructor TSortList.Create;	// *OK* Create indexed table -> Have to "Define" it before usage
	begin
	FData:=nil;				// Data
	FIndex:=nil;			// Index
	FRecNo:=1;				// To 1st record
	MaxRecordCount:=0;// Last known max. record count
	FRecordCount:=0;	// No records
	FBOF:=True;				// At Begining Of File
	FEOF:=True;				// At End Of File
	FSize:=0;					// Initial value (no meaning)
	FPtr:=Ptr(Integer(@FSize)+SizeOf(FSize));
	end;

// NonIndex = First non-indexed field in the record definition.
//						If all fields are indexed, send the End-mark field here
// EndMark = byte (or boolean) end-mark field, put it after the last record field.
//					(doesn't belong to record definition, isn't memorized)
procedure TSortList.Define(var NonIndex,EndMark);
	begin
	FSize:=Integer(@EndMark)-Integer(@FSize)-SizeOf(FSize);		// Record size (with index fields)
	FISize:=Integer(@NonIndex)-Integer(@FSize)-SizeOf(FSize);	// Index size
	if FISize<255 then FISize2:=FISize+1 else FISize2:=0;			// "Deleted" record mark
	FSize2:=FSize+1;	// Record size + 1 additional index-length byte
	ClearRecord;			// Prepare empty record
	end;

destructor TSortList.Destroy;	// *OK* Destroy indexed table
	begin
	EmptyTable;
	inherited;
	end;

procedure TSortList.ClearRecord;		// *OK* Clear record
	begin
	FillChar(FPtr^,FSize,#0);
	end;

procedure TSortList.Append;		// *OK* Append one record
	begin
	FRecNo:=FRecordCount+1;
	FBOF:=False;
	FEOF:=True;
	ClearRecord;
	end;

procedure TSortList.Edit;		// *OK* Edit record
	begin
	if FRecNo>FRecordCount then
		raise Exception.Create('Past End of File !');
	end;

procedure TSortList.First;		// *OK* To first record
	begin
	FRecNo:=1;
	FBOF:=True;
	if FRecordCount>0 then
		begin
		ReadRecord;
		FEoF:=False;
		end
	else
		FEoF:=True;
	end;

procedure TSortList.Last;			// *OK* To last record
	begin
	FEOF:=True;
	if FRecordCount>0 then
		begin
		FRecNo:=FRecordCount;
		FBoF:=False;
		ReadRecord;
		end
	else
		begin
		FRecNo:=1;
		FBoF:=True;
		end;
	end;

procedure TSortList.Next;			// *OK* Move one forward
	begin
	if FRecNo<FRecordCount then
		begin
		Inc(FRecNo);
		FBoF:=False;
		ReadRecord;
		end
	else
		FEoF:=True;
	end;

procedure TSortList.Prior;		// *OK* Move one back
	begin
	if FRecNo>1 then
		begin
		Dec(FRecNo);
		FEoF:=False;
		ReadRecord;
		end
	else
		FBoF:=True;
	end;

procedure TSortList.Move(a:integer);		// *OK* Move in any direction
	begin
	Inc(FRecNo,a);
	if FRecNo>FRecordCount then
		begin
		FRecNo:=FRecordCount;
		FEoF:=True;
		end
	else
		FEoF:=False;
	if FRecNo<1 then
		begin
		FRecNo:=1;
		FBoF:=True;
		end
	else
		FBoF:=False;
	if FRecordCount>0 then ReadRecord;
	end;

procedure TSortList.ReadRecord;				// *OK* Read record (from table to vars)
	begin
	System.Move(FData^[FIndex^[FRecNo]+1],FPtr^,FSize);
	end;

procedure TSortList.WriteRecord;			// *OK* Write record (from vars to table)
	begin
	System.Move(FPtr^,FData^[FIndex^[FRecNo]+1],FSize);
	end;

procedure TSortList.Post;			// *OK* Post data (after Append or Edit)
	begin
	if FRecNo>FRecordCount then
		begin
		if (FISize>0) and GoToNearest then	// Find the perfect place
			raise Exception.Create('Key violation !');
		InsertRecord;
		end;
	WriteRecord;
	end;

procedure TSortList.Cancel;		// *OK* Cancel editing data
	begin
	if FRecNo>FRecordCount then
		begin
		FEOF:=True;
		if FRecordCount>0 then
			begin
			FRecNo:=FRecordCount;
			ReadRecord;
			end
		else
			begin
			FRecNo:=1;
			FBoF:=True;
			end;
		end
	else
		ReadRecord;
	end;

procedure TSortList.InsertRecord;		// *OK* Insert one record (before this one)
	begin
	Inc(FRecordCount);
	if FRecordCount>MaxRecordCount then
		begin
		MaxRecordCount:=FRecordCount;
		ReAllocMem(FData,MaxRecordCount*FSize2);
		ReAllocMem(FIndex,MaxRecordCount*4);
		Temp:=(MaxRecordCount-1)*FSize2;	// Append at the end
		end
	else
		Temp:=FIndex^[FRecordCount];	// Append where last was deleted
	if FRecordCount>FRecNo then
		System.Move(FIndex^[FRecNo],FIndex^[FRecNo+1],(FRecordCount-FRecNo)*4);
	FIndex^[FRecNo]:=Temp;
	FData^[FIndex^[FRecNo]]:=FISize;
	end;

procedure TSortList.Delete;		// *OK* Delete this record
	begin
	if FRecordCount=1 then
		EmptyTable
	else
		begin
		Dec(FRecordCount);
		Temp:=FIndex^[FRecNo];
		FData^[Temp]:=FISize2;	// mark as deleted
		if FRecordCount>=FRecNo then
			System.Move(FIndex^[FRecNo+1],FIndex^[FRecNo],(FRecordCount-FRecNo+1)*4);
		FIndex^[FRecordCount+1]:=Temp;
		if FRecordCount<FRecNo then	// Past end of file ?
			begin
			FEoF:=True;
			if FRecordCount>0 then		// No more records ?
				begin
				FBoF:=False;
				FRecNo:=FRecordCount;
				end
			else
				begin
				FBoF:=True;
				FRecNo:=1;
				end;
			end
		else
			begin
			FBoF:=False;
			FEoF:=False;
			end;
		end;
	end;

procedure TSortList.EmptyTable;		// *OK* Make table Empty
	begin
	if FRecordCount>0 then
		begin
		ReAllocMem(FData,0);
		ReAllocMem(FIndex,0);
		MaxRecordCount:=0;
		FRecordCount:=0;
		FRecNo:=1;
		FBoF:=True;
		FEoF:=True;
		end;
	end;

function TSortList.Find(const s:ShortString):boolean;
	var
		LowI,HighI,NowI:integer;
		Found:boolean;
		S2:PSString;
	begin
	if FRecordCount=0 then
		begin
		Result:=False;
		FRecNo:=1;
		FBoF:=True;
		FEoF:=True;
		Exit;
		end;
	Found:=False;
	LowI:=1;
	HighI:=FRecordCount;
	repeat
		NowI:=(HighI+LowI) div 2;	// Middle
		S2:=Ptr(Integer(FData)+FIndex^[NowI]);
		if s2^>s then HighI:=NowI-1 		// Move Lower
		else if s2^<s then LowI:=NowI+1	// Move Higher
		else
			begin
			Found:=True;
			Break;
			end;
		until LowI>HighI;
	FRecNo:=NowI;
	if Found then
		begin
		FBoF:=FRecNo=1;
		FEoF:=FRecordCount=FRecNo;
		Result:=True;
		end
	else
		begin
		if s2^<s then Inc(FRecNo);
		FBoF:=FRecNo=1;
		FEoF:=FRecordCount=FRecNo;
		Result:=False;
		end;
	end;

function TSortList.GoToNearest:boolean;
	var
		S:ShortString;
	begin
	// Fill scan-string "s"
	SetLength(s,FISize);
	System.Move(FPtr^,S[1],FISize);
	// Start searching ...
	Result:=Find(s);
	end;

function TSortList.FindKey(const Values:array of const):boolean;
	var
		I:integer;
		P:byte;
		S:ShortString;
		d:double;
		OldPos:integer;
	begin
	// Fill scan-string "s"
	SetLength(s,FISize); P:=1;
	for I := 0 to High(Values) do
		with Values[I] do case VType of
			vtInteger:	// Long integer (small and short integer not supported as index fields)
				begin
				System.Move(VInteger,S[P],SizeOf(VInteger));
				Inc(P,SizeOf(VInteger));
				end;
			vtCurrency:	// Double (default Float type)
				begin
				System.Move(VCurrency,S[P],SizeOf(VCurrency^));
				Inc(P,SizeOf(VCurrency^));
				end;
			vtExtended:	// Extended type not supported, always converted to Double !
				begin
				d:=VExtended^;
				System.Move(d,S[P],SizeOf(d));
				Inc(P,SizeOf(d));
				end;
			vtChar:
				begin
				System.Move(VChar,S[P],SizeOf(VChar));
				Inc(P,SizeOf(VChar));
				end;
			vtPChar,vtAnsiString:	// Long string (only the last field in index can be of ShortString type)
				begin
				S[P]:=Char(Length(String(VPChar)));
				System.Move(String(VPChar)[1],S[P+1],Length(String(VPChar)));
				Inc(P,Length(String(VPChar))+1);
				end;
			vtString:	// Short string (only the last field in index can be of ShortString type)
				begin
				S[P]:=Char(Length(VString^));
				System.Move(VString^[1],S[P+1],Length(VString^));
				Inc(P,Length(VString^)+1);
				end;
			vtBoolean:	// Boolean not recommended as index field
				begin
				System.Move(VBoolean,S[P],SizeOf(VBoolean));
				Inc(P,SizeOf(VBoolean));
				end;
			end;
	if P<=FISize then FillChar(S[P],FISize-P+1,#0);	// Clear the rest
	// Start searching ...
	OldPos:=FRecNo;
	if Find(s) then
		begin
		FEoF:=False;
		FBoF:=False;
		ReadRecord;
		Result:=True;
		end
	else
		begin
		FRecNo:=OldPos;
		Result:=False;
		end;
	end;

end.
